// File generated automatically by build.rs. Do not edit.
#[cfg(feature = "string-module")]
mod string;
#[cfg(feature = "macho-module")]
mod macho;
#[cfg(feature = "pe-module")]
mod pe;
#[cfg(feature = "elf-module")]
mod elf;
#[cfg(feature = "text-module")]
mod text;
#[cfg(feature = "dotnet-module")]
mod dotnet;
#[cfg(feature = "lnk-module")]
mod lnk;
#[cfg(feature = "hash-module")]
mod hash;
#[cfg(feature = "magic-module")]
mod magic;
#[cfg(feature = "math-module")]
mod math;
#[cfg(feature = "test_proto2-module")]
mod test_proto2;
#[cfg(feature = "time-module")]
mod time;
#[cfg(feature = "test_proto3-module")]
mod test_proto3;
#[cfg(feature = "console-module")]
mod console;